#if defined(__GNUC__)
#  include <sys/cygwin.h>
#  define MS_EXPORT
#  define GNU_EXPORT __attribute ((dllexport))
#endif
#if defined(_MSC_VER)
#  define MS_EXPORT  __declspec (dllexport)
#  define GNU_EXPORT
void cygwin_conv_to_full_win32_path(const char *in,char *out);
#endif
#include <Windows.h>
#include <string.h>
#include <stdio.h>


/* Notes:                                                                               */
/*   All GetPath functions require caller to allocate buffers for desired paths,        */
/*      and specify the size of the output buffer with 2nd parameter, buff_size.        */
/*   Path requested is valid                      if ...   return_value > 0             */
/*   Path requested is returned in output buffer  if ...   0 < return_value < buff_size */

MS_EXPORT size_t GetShortPath    (const char *path,const size_t buff_size,char *shortpath) GNU_EXPORT;
MS_EXPORT size_t GetLongPath     (const char *path,const size_t buff_size,char *longpath)  GNU_EXPORT;
MS_EXPORT size_t GetShortPath_nt4(const char *path,const size_t buff_size,char *shortpath) GNU_EXPORT;
MS_EXPORT size_t GetLongPath_nt4 (const char *path,const size_t buff_size,char *longpath)  GNU_EXPORT;
size_t RemoveQuotes    (const char *in,const size_t buff_size,char *out);        /* For internal use only */

size_t GetShortPath_nt4(const char *longpath,const size_t buff_size,char *shortpath) {
  /* GetShortPathName is broken on Win NT4 */
  /* FindFirstFile does not append the directory/drive info, so */
  /* we must search one directory at a time to get short dir names */
  /* as well as the short form of the file name.*/
  size_t valid = 1;
  char   findtemp[MAX_PATH];
  char   shorttemp[MAX_PATH];
  size_t short_len;
  char   *begin,*end;

  /* Remove any surrounding quotes. */
  RemoveQuotes(longpath,MAX_PATH,shorttemp);

  /* cygwin preserves short/long'ness of paths */
  cygwin_conv_to_full_win32_path(shorttemp,findtemp); /* cygwin.h */

  /* Initialize */
  /* cygwin sets findtemp to <Drive>:\<something> */
  /* So, we should skip the first <Drive>:\ part which FindFirstFile doesn't handle. */
  short_len = 3;
  strncpy(shorttemp,findtemp,short_len);
  shorttemp[short_len] = '\0';
  begin = findtemp + short_len;
  end   = begin;

  while(end) {
    WIN32_FIND_DATA path;              /* Windows.h */
    /* End the search string at the next "\" */
    end = strchr(begin,(int)'\\');
    if (end) *end = '\0';
    if (INVALID_HANDLE_VALUE != FindFirstFile(findtemp,&path)) { /* Windows.h */
      /* FindFirstFile returns the short form of the last part of the path in path.cAlternateFileName */
      /* ONLY if it is different from the long form which is in path.cFileName */
      if (path.cAlternateFileName[0]!='\0') {
        strcat(shorttemp,path.cAlternateFileName); /* Windows.h */
      } else {
        strcat(shorttemp,path.cFileName); /* Windows.h */
      }
      /* Restore the removed "\" and add one to longtemp if necessary */
      if (end) {
        strcat(shorttemp,"\\");
        *end  = '\\';
        /* Skip over this "\" for the next call to FindFirstFile */
        end++;
        begin = end;
      }
    } else {
      /* File does not exist! */
      valid = 0;
      break;
    }
  }

  short_len = strlen(shorttemp);
  if (valid) {
    /* Only fill the output buffer if sufficient size exists to hold the result */
    if (buff_size>=short_len) {
      strcpy(shortpath,shorttemp);
    }
    /* If file exists, ALWAYS return the required buffer length. */
    valid = short_len;
  }
  return(valid);
}

size_t GetLongPath_nt4(const char *shortpath,const size_t buff_size,char *longpath) {
  /* GetLongPathName does not exist on Win NT4, so we must use FindFirstFile. */
  /* FindFirstFile does not append the directory/drive info, so */
  /* we must search one directory at a time to get long dir names */
  /* as well as the long form of the file name.*/
  char   longtemp[MAX_PATH]; /* Windows.h */
  char   findtemp[MAX_PATH]; /* Windows.h */
  size_t valid  = 1;
  size_t long_len;
  char   *begin,*end;

  RemoveQuotes(shortpath,MAX_PATH,longtemp);

  /* cygwin preserves short/long'ness of paths */
  cygwin_conv_to_full_win32_path(longtemp,findtemp); /* cygwin.h */

  /* Initialize */
  /* cygwin sets findtemp to <Drive>:\<something> */
  /* So, we should skip the first <Drive>:\ part which FindFirstFile doesn't handle. */
  long_len=3; 
  strncpy(longtemp,findtemp,long_len);
  longtemp[long_len] = '\0';
  begin = findtemp + long_len;
  end   = begin;

  while (end) {
    WIN32_FIND_DATA path;              /* Windows.h */
    /* End the search string at the next "\" */
    end = strchr(begin,(int)'\\');
    if (end) *end = '\0';
    /* FindFirstFile returns the long form of the last part of the path in path.cFileName */
    if (INVALID_HANDLE_VALUE != FindFirstFile(findtemp,&path)) { /* Windows.h */
      strcat(longtemp,path.cFileName); /* Windows.h */
      /* Restore the removed "\" and add one to longtemp if necessary */
      if (end) {
        strcat(longtemp,"\\");
        *end  = '\\';
        /* Skip over this "\" for the next call to FindFirstFile */
        end++;
        begin = end;
      }
    } else {
      /* File does not exist! */
      valid = 0;
      break;
    }
  }

  long_len = strlen(longtemp);
  if (valid) {
    /* Only fill the output buffer if sufficient size exists to hold the result */
    if (buff_size>=long_len) {
      strcpy(longpath,longtemp);
    }
    /* If file exists, ALWAYS return the required buffer length. */
    valid = long_len;
  }
  return(valid);
}

size_t GetShortPath(const char *longpath,const size_t buff_size,char *shortpath) {
  char   tmppath[MAX_PATH];    /* Windows.h */
  char   short_path[MAX_PATH]; /* Windows.h */
  DWORD  buffsize;

  RemoveQuotes(longpath,MAX_PATH,short_path);

  /* cygwin preserves short/long'ness of paths */
  cygwin_conv_to_full_win32_path(short_path,tmppath); /* cygwin.h */

  /* buffsize = 0 denotes failure of GetShortPathName */
  buffsize = GetShortPathName(tmppath,short_path,MAX_PATH); /* Windows.h */
  /* Only fill the output buffer if sufficient size exists to hold the result */
  if ((buffsize>0) && (buffsize<buff_size)) {
    strcpy(shortpath,short_path);
  }
  /* If file exists, ALWAYS return the required buffer length. */
  return(buffsize);
}

size_t GetLongPath(const char *shortpath,const size_t buff_size,char *longpath) {
  char   tmppath[MAX_PATH];    /* Windows.h */
  char   long_path[MAX_PATH]; /* Windows.h */
  DWORD  buffsize;

  RemoveQuotes(shortpath,MAX_PATH,long_path);
  /* cygwin preserves short/long'ness of paths */
  cygwin_conv_to_full_win32_path(long_path,tmppath); /* cygwin.h */

  /* buffsize = 0 denotes failure of GetLongPathName */
  buffsize = GetLongPathName(tmppath,long_path,MAX_PATH); /* Windows.h */

  /* Only fill the output buffer if sufficient size exists to hold the result */
  if (buffsize>0 && buffsize<buff_size) {
    strcpy(longpath,long_path);
  }
  /* If file exists, ALWAYS return the required buffer length. */
  return(buffsize);
}

size_t RemoveQuotes(const char *in,const size_t buff_size,char *out) {
  const char *str=in;
  size_t len;

  len = strlen(in);
  if ((in[0]=='\'')||(in[0]=='\"')) {
    str++;
    len--;
  }
  if ((str[len-1]=='\'')||(str[len-1]=='\"')) {
    len--;
  }
  if (buff_size>=len) {
    strncpy(out,str,len);
    out[len]='\0';
  }
  return(len);
}

#if defined(_MSC_VER)
typedef void (*my_cygwin_file_api)(const char *,char *);
void cygwin_conv_to_full_win32_path(const char *in,char *out) {
  HMODULE cygwrapper = LoadLibrary("mycygwrapper.dll");
  if (cygwrapper) {
    my_cygwin_file_api cfwp;
    cfwp = (my_cygwin_file_api)GetProcAddress(cygwrapper,"my_cygwin_conv_to_full_win32_path");
    if (cfwp) {
      cfwp(in,out);
    }
    FreeLibrary(cygwrapper);
  }
}
#endif
